! Triplet.f90 computes triplet powder spectra for given D, E, and g-matrix.
! The lineshape is Gaussian or Lorentzian, whose width is from g-strain or from field broadening.
! D, and E can be distributed.
! First-order hyperfine interaction from two equivalent nuclei is included.
    
! WRHagen @ 20230828
    
!   ===================================================================================    
    MODULE SimulationParameters
!   joint parameters between Triplet (main) and FieldLineShape (subroutine)
!   wrh 20151003
    IMPLICIT NONE
    SAVE
    
    REAL    ::  bmin, bmax, width, bval, perp, parl, gstrain 
    INTEGER ::  ksim, ntrans, l, jhyp
    REAL    ::  t1parl(1024), t2parl(1024), t3parl(1024), tdparl(1024)
    REAL    ::  t1perp(1024), t2perp(1024), t3perp(1024), tdperp(10240)
    
!   t1 is 1-->2 (deltaM = 1)
!   t2 is 1-->3 (deltaM = 2)
!   t3 is 2-->3 (deltaM = 1)
!   td is 1-->3 (deltaM = 1+1)

    END MODULE SimulationParameters
!   ===================================================================================
!
    subroutine Triplet (simpar,simint,a1,a2,a3,a4,e1,e2,e3,e4)

  !DEC$ ATTRIBUTES DLLEXPORT::triplet
  !DEC$ ATTRIBUTES STDCALL::triplet
    
    USE SimulationParameters
    
    IMPLICIT NONE

  ! Variables

	REAL(8)	:: machep
	REAL(8)	:: emat(3,3), zmat(3,3), areal(3,3), aimgy(3,3)
	REAL(8)	:: diago(3), offdi(3), dummy(3), tau(3,3)
	REAL(8)	:: vecre(3,3), vecim(3,3)
	COMPLEX	:: vcmplx(3,3)
	REAL	:: bb(32),en(32,3),en2(32,3),er(1024,3)
	REAL	:: edum(32),edum2(32),bvalM(3,3,2),perpM(3,3,2),parlM(3,3,2)
	REAL(8)	:: emaL(3,3),emaR(3,3),zdum(3,3)
	REAL	:: gstrM(3,3,2)
    
    REAL(8) :: wval,wz,wy,wx,gammi

    REAL    :: simpar(17)
    INTEGER :: simint(4),iGauss,iGwidth,i,nlz,nphi,nsigma,ii,ij,ik,j,k,m,j1,j2,ierr,kv,lv,ihyp
    REAL    :: a1(1024),a2(1024),a3(1024),a4(1024),e1(1024),e2(1024),e3(1024),e4(1024)
    REAL    :: sparl(1024),sperp(1024)
    REAL	:: freq,hanu,brange,bstep,elen2,cut,b,energy,eprev,energyl,energyr
    REAL	:: zl,yl,xl,stepzl,phinum,stepphi,sinthe,phi
    REAL	:: gz,gy,gx,dd,ee,az,ay,ax,sz,sy,sx,gamma,gg,aval,rhyp,phyp
    REAL	:: d,e,dsigma,esigma,shiftdsigma,shiftesigma
    REAL	:: rangedsigma,rangeesigma,stepdsigma,stepesigma,shiftsigma,sigmafactord, sigmafactore
	

	! All distributions are Gaussians:
	! F(x) = (1/s) Fmax EXP [-ln2((x-x0)/s)**2]
	! in which linewidth s is HWHH (half width at half height]

    DO i=1,1024
        a1(i)=0.;a2(i)=0.;a3(i)=0.;a4(i)=0.
        e1(i)=0.;e2(i)=0.;e3(i)=0.;e4(i)=0.
        t1parl(i)=0.;t2parl(i)=0.;t3parl(i)=0.;tdparl(i)=0.
        t1perp(i)=0.;t2perp(i)=0.;t3perp(i)=0.;tdperp(i)=0.
    END DO

!	Re-write the input parameters:
	freq	= simpar(1)		! [MHz]
	gz		= simpar(2)		! [dim.less]
	gy		= simpar(3)
	gx		= simpar(4)
	dd		= simpar(5)		! [cm-1]
	ee		= simpar(6)		! [cm-1]
    iGauss	= simint(1)				! If pos than Gaussianl if neg then Lorentzian
	nlz		= ABS(simint(1))		! steps in cos(theta)
    iGwidth	= simint(2)				! If pos than g-strain; if neg then gauss width
	nphi	= ABS(simint(2))		! steps in phi
	gammi	= simpar(7)		! [Gauss] Gaussian standard deviation of B-distribution
	Dsigma	= simpar(8)		! [cm-1] Gaussian HWHH of ZFS-distribution
	nsigma	= simint(3)		! steps in ZFS distribution
	Esigma	= simpar(9)		! as Dsigma
	ksim	= simint(4)		! number of spectral points
	bmin	= simpar(10)	! [Gauss] Start field of spectrum
	bmax	= simpar(11)	! [Gauss] End field

	az		= simpar(12)	! [Gauss] 1st order hyperfine from I=3/2+3/2
	ay		= simpar(13)
	ax		= simpar(14)

    ! [g-value units] diagonal g-strain (standard deviation)
    IF (iGwidth .GT. 0) THEN
		sZ	= simpar(15)
		sY	= simpar(16)
		sX	= simpar(17)
        wz	= 0.
        wy	= 0.
        wx	= 0.
    ELSE
        sZ	= 0.
        sY	= 0.
        sX	= 0.
		wz		= simpar(15)	
		wy		= simpar(16)
		wx		= simpar(17)
    END IF

	brange	= bmax - bmin
	bstep	= brange / (REAL(ksim)-1.)

!	Do some safety checking:
	IF (nphi.LE.0) nphi = 1
	IF (gammi.LE.0.) gammi	= bstep/10. 

!	The microwave energy:
	hanu	= (0.33361E-4) * freq

!	Parameters for use in Gaussian distribution:
	elen2	= - ALOG (2.)
	cut		= SQRT ( ALOG (0.001) / elen2 )		! 3.16 standard deviations

!	What is the smallest number on this computer (for matrix diagonalization):
	CALL machine_precision (machep)

!	dd, ee are (my) universal parameters for comparison with S = 2, 3, 4, etc.
	d		= dd / 3.
	e		= ee
    shiftDsigma = 0.0
    shiftEsigma = 0.0

!	Initialize (steps in) cos(theta) and phi:
	zl		= 1.0					!For theta=0 cos(theta)=1
	stepzl	= 0.
	IF (nlz.GT.1) stepzl = 1.0/(REAL(nlz)-1.)
	zl		= zl + stepzl
	phinum	= 1.570796
	stepphi	= 0.

	IF (nphi.GT.1) stepphi = phinum/(REAL(nphi)-1.)

!	Initialize (steps in) ZFS distribution:
	IF (nsigma .GT. 1) THEN
		IF (Dsigma .EQ. 0.) Dsigma = 0.00001
		rangeDsigma = cut * Dsigma	!Go out 3.15 st.devs to 0.1% intensity
		rangeEsigma = cut * Esigma
		stepDsigma = 2.*rangeDsigma / (REAL(nsigma)-1.)
		stepEsigma = 2.*rangeEsigma / (REAL(nsigma)-1.)
	END IF
	shiftsigma  = 0.
	sigmafactorD = 1.
    sigmafactorE = 1.

!	Here starts the [cos(theta),phi] scan over the unit sphere:
	DO  ii = 1, nlz							! step in cos(theta)
	    zl = zl - stepzl

	DO	ij = 1, nphi
	    sinthe = SQRT(1. - zl*zl)
	    phi = phinum - (ij-1) * stepphi
	    yl = sinthe * SIN(phi)
	    xl = sinthe * COS(phi)

!	Lorentz anisotropic linewidth:	gamma = HWHH
    wval= xl*xl*wx*wx + yl*yl*wy*wy +zl*zl*wz*wz
    gamma = gammi*gammi + wval
    gamma = SQRT(gamma)
    
!	A first-order hyperfine structure (I=3/2+3/2):
	gg	= xl*xl*gx*gx + yl*yl*gy*gy + zl*zl*gz*gz
	aval= xl*xl*gx*gx*gx*gx*ax*ax
	aval= yl*yl*gy*gy*gy*gy*ay*ay + aval
	aval= zl*zl*gz*gz*gz*gz*az*az + aval
	aval= (SQRT(aval))/gg

!	Here starts the ZFS-distribution loop:
!	Note that distributed VERSUS undistributed spectra are NOT normalized, i.e.
!	amplitudes of spectra with sigma.NE.0 cannot be compared with sigma=0 spectra 
	    DO	ik = 1, nsigma					! step in zf splittings
		IF (nsigma .GT. 1) THEN
            IF (Dsigma .NE. 0.) THEN									! 10 lines modified 20210220
				shiftDsigma  = rangeDsigma - (ik-1) * stepDsigma
				sigmafactorD = elen2 * (( shiftDsigma / Dsigma )**2)
				sigmafactorD = EXP (sigmafactorD)
            END IF
            ! NOTE that the D and E strain are here fully correlated
            !(i.e. NOT two nested DO-loops).
            IF (Esigma .NE. 0.) THEN
				shiftEsigma	 = rangeEsigma - (ik-1) * stepEsigma
				sigmafactorE = elen2 * (( shiftEsigma / Esigma )**2)
				sigmafactorE = EXP (sigmafactorE)
            END IF
		END IF

        d = (dd/3.) + shiftDsigma
        e = ee + shiftEsigma

!	Re-set the resonance fields and transition probabilities:
		FORALL (i=1:3,j=1:3,l=1:2)
			bvalM(i,j,l) = 0.
			perpM(i,j,l) = 0.
			parlM(i,j,l) = 0.
		END FORALL

!	Construct the field-independent matrices emat and zmat (with g-strain):
		CALL PREM2H (gz-sZ,gy-sY,gx-sX,d,e,zl,yl,xl,emaL,zdum)
		CALL PREM2H (gz,gy,gx,d,e,zl,yl,xl,emat,zmat)
		CALL PREM2H (gz+sZ,gy+sY,gx+sX,d,e,zl,yl,xl,emaR,zdum)

!	Exactly calculate a 32*3 grid of energy levels over the field range:
		k = 32
        DO      i = 1, k
                bb(i) = bmin + (i-1) * ( brange / (k-1.) )
                CALL DIAM2H (bb(i),emat,zmat,areal,aimgy)
                DO      j1 = 1, 3
                        DO      j2 = 1, 3
                                vecre(j1,j2) = 0.
                                vecim(j1,j2) = 0.
                        END DO
                        vecre(j1,j1) = 1.
                END DO
                CALL HTRIDI (3,3,areal,aimgy,diago,offdi,dummy,tau)
                CALL TQL2 (machep,3,3,diago,offdi,vecre,ierr)
                DO      j = 1, 3
                        en(i,j) = diago(j)
                END DO
        END DO

!	Spline-interpolate ksim * 3 energy levels over the field range:
        DO      j = 1, 3
                DO      i = 1, k
                        edum(i) = en(i,j)
                END DO
                CALL SPLIN0 (bb,edum,k,edum2)						! construct table (dim of bb <= 50)
                DO      i = 1, k
                        en2(i,j) = edum2(i)
                END DO
        END DO
        DO      j = 1, 3
                DO      i = 1, k
                        edum (i) = en (i,j)
                        edum2(i) = en2(i,j)
                END DO
                DO      m = 1, ksim
                        b = bmin + (m-1.) * bstep
                        CALL SPLINT (bb,edum,edum2,k,b,energy)		! call table
                        er (m,j) = energy
                END DO
        END DO

!	Compute resonance conditions at fields bvalM for which Ej-Ei-hanu=0 :
        DO  j = 2, 3
            DO  i = 1, j-1
                eprev = er(1,j) - er(1,i) - hanu + 2. * machep
                l     = 1
                DO      m = 2, ksim
                        IF (l .GT. 2) EXIT
                        energy = er(m,j) - er(m,i) - hanu
                        IF ( (ABS(energy).LE.1.E-6).OR.((energy / eprev).LT.0.) ) THEN
                                bvalM(i,j,l) = bmin+(m-1.5)*bstep
                                l = l + 1
                        END IF
                        eprev = energy + 2. * machep
                END DO
            END DO
        END DO

!	Calculate parlM and perpM transition probs at resonance fields bvalM(i,l):
        DO  j = 2, 3
        DO  i = 1, j-1
            DO  l = 1,2
                IF ( bvalM(i,j,l) .EQ. 0.0) EXIT

                CALL DIAM2H (bvalM(i,j,l),emat,zmat,areal,aimgy)
                DO  j1 = 1, 3
                    DO  j2 = 1, 3
                        vecre(j1,j2) = 0.
                        vecim(j1,j2) = 0.
                    END DO
                    vecre(j1,j1) = 1.
                END DO
                CALL HTRIDI (3,3,areal,aimgy,diago,offdi,dummy,tau)
                CALL TQL2   (machep,3,3,diago,offdi,vecre,ierr)
                CALL HTRIBK (3,3,areal,aimgy,tau,3,vecre,vecim)
                DO  kv = 1, 3
                    DO  lv = 1, 3
                    vcmplx(kv,lv)=(vecre(kv,lv)+(0.,1.)*vecim(kv,lv))
                    END DO
                END DO
                CALL INTNS2H (i,j,xl,yl,zl,perp,parl,gx,gy,gz,vcmplx)
                perpM(i,j,l) = perp
                parlM(i,j,l) = parl

!			Now calculate g-strain from left and from right of Bres:
				CALL DIAM2H (bvalM(i,j,l),emaL,zmat,areal,aimgy)
                DO  j1 = 1, 3
                    DO  j2 = 1, 3
                        vecre(j1,j2) = 0.
                        vecim(j1,j2) = 0.
                    END DO
                    vecre(j1,j1) = 1.
                END DO
                CALL HTRIDI (3,3,areal,aimgy,diago,offdi,dummy,tau)
                CALL TQL2   (machep,3,3,diago,offdi,vecre,ierr)
				energyL = diago(j)-diago(i)
				CALL DIAM2H (bvalM(i,j,l),emaR,zmat,areal,aimgy)
                DO  j1 = 1, 3
                    DO  j2 = 1, 3
                        vecre(j1,j2) = 0.
                        vecim(j1,j2) = 0.
                    END DO
                    vecre(j1,j1) = 1.
                END DO
                CALL HTRIDI (3,3,areal,aimgy,diago,offdi,dummy,tau)
                CALL TQL2   (machep,3,3,diago,offdi,vecre,ierr)
				energyR = diago(j)-diago(i)
				gstrM(i,j,l) = (ABS (energyL-energyR)) / ((SQRT(gg))*(0.466882E-4))
				gstrM(i,j,l) = gstrM(i,j,l) / 2. ! now it is a Gaussian standard deviation
!			End of g-strain line width calculation

            END DO
        END DO
        END DO

!	Now put Gaussians on the stick spectrum:
		DO	j = 2,3
			DO	i = 1, j-1
!	Count transition (#=1,2,3) for the calculation of the double-quantum transition:
				ntrans = j - 2 + i

				DO	l = 1, 2
					bval = bvalM(i,j,l)
					perp = perpM(i,j,l) * sigmafactorD * sigmafactorE
					parl = parlM(i,j,l) * sigmafactorD * sigmafactorE
					width= SQRT(gstrM(i,j,l)*gstrM(i,j,l)+gamma*gamma)
                    gstrain=gstrM(i,j,l)
					jhyp = 1
                    IF (aval .GT. 0.) THEN
						DO	ihyp = -3,3,1
							jhyp = ihyp + 4
							rhyp = REAL(ihyp)
							bval = bvalM(i,j,l) + rhyp*aval
							IF (bval .LE. 0.) CYCLE
							phyp = (4.-ABS(rhyp))/4.
							perp = perpM(i,j,l) * phyp * sigmafactorD * sigmafactorE
							parl = parlM(i,j,l) * phyp * sigmafactorD * sigmafactorE					
							CALL FieldLineShape(iGauss)						
						END DO
                    ELSE
                        CALL FieldLineShape(iGauss)
                    END IF
 				END DO
			END DO
		END DO


	    END DO			! zfs distr. run ends

	END DO				! phi loop ends
    
	END DO				! cos-theta loop ends

    DO  i = 1, ksim
        a1(i) = t1parl(i)
        a2(i) = t2parl(i)
        a3(i) = t3parl(i)
        a4(i) = tdparl(i)
        e1(i) = t1perp(i)
        e2(i) = t2perp(i)
        e3(i) = t3perp(i)
        e4(i) = tdperp(i)
    END DO        

	RETURN
	
    end subroutine Triplet

! @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
	SUBROUTINE FieldLineShape(iGauss)

!	WRHagen 20030410

    USE SimulationParameters

	REAL	:: dqfact(1024,7)				! expon factor
    INTEGER	:: iGauss

	elen2	= - ALOG (2.)
    IF (iGauss .GT. 0) THEN
		bcut = SQRT ( ALOG (0.0001) / elen2 )	! Gaussian
    ELSE
		bcut = 16.	! Lorentzian
    END IF
	bstep	= (bmax-bmin) / (ksim - 1.)

	rangeb	= bcut * width
	enb	= 2. * rangeb / bstep + 0.5
	nb	= enb + 1

	bstart	= bval - rangeb
	nstart	= 0.5 + (bstart-bmin)/bstep
	bstart	= bmin + (nstart-1) * bstep

	IF (ntrans.EQ.1 .AND. l.EQ.1) THEN
		DO	j = 1, 1024
			dqfact(j,jhyp) = 0.0
		END DO
	END IF
						
	DO	ib = 1, nb
		bsweep	= bstart + (ib-1) * bstep
		IF (bsweep .LT. bmin) CYCLE
		IF (bsweep .GT. bmax) GO TO 2

		bsweep	= bsweep - bval
        
        IF (iGauss .GT. 0) THEN
  			expon	= -0.5 * (bsweep / width) **2
 			expon	= EXP (expon)
        ELSE
			expon	= width**2
			expon	= expon / (expon + bsweep**2)
        END IF

        expon = expon / width       ! normalization
        
		i	= nstart + ib - 1
		IF (i .LT. 1) CYCLE
		IF (i .GT. ksim) GO TO 2
        
        SELECT CASE (ntrans)
        CASE (1)
				dqfact(i,jhyp) = perp*expon
                t1parl(i) = t1parl(i) + parl*expon
                t1perp(i) = t1perp(i) + perp*expon
            CASE (2)
                t2parl(i) = t2parl(i) + parl*expon
                t2perp(i) = t2perp(i) + perp*expon
            CASE (3)
                t3parl(i) = t3parl(i) + parl*expon
                t3perp(i) = t3perp(i) + perp*expon
                tdperp(i) = tdperp(i) + (dqfact(i,jhyp)*perp*expon*width*width)*expon
                ! Product probability: perp(ntrans=1) * perp(ntrans=3), times a width-normalized (ntrans=3)
                ! it has two unit intensities and one width-normalized intensity.
            CASE DEFAULT
				CONTINUE
            END SELECT

	END DO
 2	CONTINUE

	RETURN
	END SUBROUTINE FieldLineShape

! @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ All following subroutines: legacy code @ @ @ @ @ @

! = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
	SUBROUTINE machine_precision (x)

! @ W.R.Hagen  891018
! This subroutine computes a floating-point precision criterion
! to be used in SUBROUTINE TQL2 (matrix diagonalization).

	REAL*8 x

	x = 1.
	DO      i = 1, 100
		x = x / 2.
		a = 1 + x
		IF (a .EQ. 1.) RETURN
	END DO
	RETURN
	END
!
!
!
! = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
	SUBROUTINE PREM2H (gz,gy,gx,d,e,zl,yl,xl,emat,zmat)
!
! @ W.R.Hagen 970301
	
	REAL*8  emat(3,3), zmat(3,3)

	bet     = 0.466882E-4
	sq2     = 1.4142136 / 2.

	DO 10   i = 1, 3
	DO 10   j = 1, 3
		zmat(i,j) = 0.
 10             emat(i,j) = 0.

! Unit-field Zeemman factors:
	zg      = gz * bet * zl
	yg      = gy * bet * yl
	xg      = gx * bet * xl

! Real zero-field elements:
	zmat (1,1) = d
	zmat (2,2) = d
	zmat (3,3) = -2.*d
	zmat (2,1) = e

! Real unit-field Zeeman elements:
	emat (1,1) =  zg
	emat (3,1) =  xg * sq2
	emat (2,2) = -zg
	emat (3,2) =  xg * sq2

! Complex unit-field Zeeman elemets:
	emat (1,3) =  yg * sq2
	emat (2,3) = -yg * sq2

	RETURN
	END
!
!
!
! = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
	SUBROUTINE DIAM2H (bval,emat,zmat,areal,aimgy)
!
! @ W.R.Hagen 970301
!
	REAL*8  emat(3,3),zmat(3,3),areal(3,3),aimgy(3,3)

	DO 10   i = 1, 3
	DO 10   j = 1, i
		areal (i,j) = emat (i,j)
 10             aimgy (i,j) = 0.

	areal (1,1) = areal (1,1) * bval
	areal (3,1) = areal (3,1) * bval
	areal (2,2) = areal (2,2) * bval
	areal (3,2) = areal (3,2) * bval

	DO 20   i = 1, 3
	DO 20   j = 1, i
		areal (i,j) = areal (i,j) + zmat (i,j)
 20     CONTINUE

	aimgy (3,1) = emat (1,3) * bval
	aimgy (3,2) = emat (2,3) * bval

	RETURN
	END
!
!
!
! = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
	SUBROUTINE intns2H (i,j,xl,yl,zl,perp,parl,gx,gy,gz,cm)
!
! @ W.R.Hagen 970301; 970702
!
	COMPLEX cm(3,3),smin,spls,sz,sy,sx

	s2      = 1.4142136

	smin =  s2*CONJG(cm(3,i))*cm(1,j) + s2*CONJG(cm(2,i))*cm(3,j)

	spls =  s2*CONJG(cm(1,i))*cm(3,j) + s2*CONJG(cm(3,i))*cm(2,j)

	sz =       CONJG(cm(1,i))*cm(1,j) -    CONJG(cm(2,i))*cm(2,j)

	sx =    (spls + smin) / 2.
	sy =    (spls - smin) / 2. / (0., 1.)

	sx =    sx * gx
	sy =    sy * gy
	sz =    sz * gz

	CALL intens (sx,sy,sz,xl,yl,zl,perp,parl)

	RETURN
	END
!
!
!
! = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
	SUBROUTINE intens (sx,sy,sz,xl,yl,zl,perp,parl)
!
! @ W.R.Hagen 930926: intens contains the part common to all
! intensity calculating subroutines intnsN.
!
	COMPLEX sx,sy,sz,ccsx,ccsy,ccsz,cparl,cperp

	ccsx =  CONJG (sx)
	ccsy =  CONJG (sy)
	ccsz =  CONJG (sz)

	cparl =         xl * xl  *  ccsx * sx
	cparl = cparl + yl * yl  *  ccsy * sy
	cparl = cparl + zl * zl  *  ccsz * sz
	cparl = cparl + xl * yl  * (ccsx * sy + ccsy * sx)
	cparl = cparl + xl * zl  * (ccsx * sz + ccsz * sx)
	cparl = cparl + yl * zl  * (ccsy * sz + ccsz * sy)

	cperp =         (1. - xl * xl) *  ccsx * sx
	cperp = cperp + (1. - yl * yl) *  ccsy * sy
	cperp = cperp + (1. - zl * zl) *  ccsz * sz
	cperp = cperp - xl * yl  * (ccsx * sy + ccsy * sx)
	cperp = cperp - xl * zl  * (ccsx * sz + ccsz * sx)
	cperp = cperp - yl * zl  * (ccsy * sz + ccsz * sy)

	perp =  REAL (cperp)
	perp =  perp / 2.
	perp =  ABS  ( perp)

	parl =  REAL (cparl)
	parl =  ABS  ( parl)

	RETURN
	END
!
!
!
! = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
      SUBROUTINE HTRIBK(NM,N,AR,AI,TAU,M,ZR,ZI)
!
      INTEGER I,J,K,L,M,N,NM
      REAL*8 AR(NM,N),AI(NM,N),TAU(2,N),ZR(NM,M),ZI(NM,M)
      REAL*8 H,S,SI
!
      IF (M .EQ. 0) GO TO 200
!
      DO 50 K = 1, N
!
	 DO 50 J = 1, M
	    ZI(K,J) = -ZR(K,J) * TAU(2,K)
	    ZR(K,J) = ZR(K,J) * TAU(1,K)
   50 CONTINUE
!
      IF (N .EQ. 1) GO TO 200
!
      DO 140 I = 2, N
	 L = I - 1
	 H = AI(I,I)
	 IF (H .EQ. 0.0D0) GO TO 140
!
	 DO 130 J = 1, M
	    S = 0.0D0
	    SI = 0.0D0
!
	    DO 110 K = 1, L
	       S = S + AR(I,K) * ZR(K,J) - AI(I,K) * ZI(K,J)
	       SI = SI + AR(I,K) * ZI(K,J) + AI(I,K) * ZR(K,J)
  110       CONTINUE
!
	    S = (S / H) / H
	    SI = (SI / H) / H
!
	    DO 120 K = 1, L
	       ZR(K,J) = ZR(K,J) - S * AR(I,K) - SI * AI(I,K)
	       ZI(K,J) = ZI(K,J) - SI * AR(I,K) + S * AI(I,K)
  120       CONTINUE
!
  130    CONTINUE
!
  140 CONTINUE
!
  200 RETURN
      END
!
!
!
! = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
      SUBROUTINE HTRIDI(NM,N,AR,AI,D,E,E2,TAU)
!
      INTEGER I,J,K,L,N,II,NM,JP1
      REAL*8 AR(NM,N),AI(NM,N),D(N),E(N),E2(N),TAU(2,N)
      REAL*8 F,G,H,FI,GI,HH,SI,SCALE
      REAL*8 DSQRT,DABS
!
      TAU(1,N) = 1.0D0
      TAU(2,N) = 0.0D0
!
      DO 100 I = 1, N
  100 D(I) = AR(I,I)
!
      DO 300 II = 1, N
	 I = N + 1 - II
	 L = I - 1
	 H = 0.0D0
	 SCALE = 0.0D0
	 IF (L .LT. 1) GO TO 130
!
	 DO 120 K = 1, L
  120    SCALE = SCALE + DABS(AR(I,K)) + DABS(AI(I,K))
!
	 IF (SCALE .NE. 0.0D0) GO TO 140
	 TAU(1,L) = 1.0D0
	 TAU(2,L) = 0.0D0
  130    E(I) = 0.0D0
	 E2(I) = 0.0D0
	 GO TO 290
!
  140    DO 150 K = 1, L
	    AR(I,K) = AR(I,K) / SCALE
	    AI(I,K) = AI(I,K) / SCALE
	    H = H + AR(I,K) * AR(I,K) + AI(I,K) * AI(I,K)
  150    CONTINUE
!
	 E2(I) = SCALE * SCALE * H
	 G = DSQRT(H)
	 E(I) = SCALE * G
	 F = DSQRT(AR(I,L)**2+AI(I,L)**2)
!
	 IF (F .EQ. 0.0D0) GO TO 160
	 TAU(1,L) = (AI(I,L) * TAU(2,I) - AR(I,L) * TAU(1,I)) / F
	 SI = (AR(I,L) * TAU(2,I) + AI(I,L) * TAU(1,I)) / F
	 H = H + F * G
	 G = 1.0D0 + G / F
	 AR(I,L) = G * AR(I,L)
	 AI(I,L) = G * AI(I,L)
	 IF (L .EQ. 1) GO TO 270
	 GO TO 170
  160    TAU(1,L) = -TAU(1,I)
	 SI = TAU(2,I)
	 AR(I,L) = G
  170    F = 0.0D0
!
	 DO 240 J = 1, L
	    G = 0.0D0
	    GI = 0.0D0
!
	    DO 180 K = 1, J
	       G = G + AR(J,K) * AR(I,K) + AI(J,K) * AI(I,K)
	       GI = GI - AR(J,K) * AI(I,K) + AI(J,K) * AR(I,K)
  180       CONTINUE
!
	    JP1 = J + 1
	    IF (L .LT. JP1) GO TO 220
!
	    DO 200 K = JP1, L
	       G = G + AR(K,J) * AR(I,K) - AI(K,J) * AI(I,K)
	       GI = GI - AR(K,J) * AI(I,K) - AI(K,J) * AR(I,K)
  200       CONTINUE
!
  220       E(J) = G / H
	    TAU(2,J) = GI / H
	    F = F + E(J) * AR(I,J) - TAU(2,J) * AI(I,J)
  240    CONTINUE
!
	 HH = F / (H + H)
!
	 DO 260 J = 1, L
	    F = AR(I,J)
	    G = E(J) - HH * F
	    E(J) = G
	    FI = -AI(I,J)
	    GI = TAU(2,J) - HH * FI
	    TAU(2,J) = -GI
!
	    DO 260 K = 1, J
	       AR(J,K) = AR(J,K) - F * E(K) - G * AR(I,K) + FI * TAU(2,K) + GI * AI(I,K)
	       AI(J,K) = AI(J,K) - F * TAU(2,K) - G * AI(I,K) - FI * E(K) - GI * AR(I,K)
  260    CONTINUE
!
  270    DO 280 K = 1, L
	    AR(I,K) = SCALE * AR(I,K)
	    AI(I,K) = SCALE * AI(I,K)
  280    CONTINUE
!
	 TAU(2,L) = -SI
  290    HH = D(I)
	 D(I) = AR(I,I)
	 AR(I,I) = HH
	 AI(I,I) = SCALE * DSQRT(H)
  300 CONTINUE
!
      RETURN
      END
!
!
!
! = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
      SUBROUTINE TQL2(machep,NM,N,D,E,Z,IERR)
!
      INTEGER I,J,K,L,M,N,II,L1,NM,MML,IERR
      REAL*8 D(N),E(N),Z(NM,N)
      REAL*8 B,C,F,G,H,P,R,S,MACHEP
      REAL*8 DSQRT,DABS,DSIGN
!
      IERR = 0
      IF (N .EQ. 1) GO TO 1001
!
      DO 100 I = 2, N
  100 E(I-1) = E(I)
!
      F = 0.0D0
      B = 0.0D0
      E(N) = 0.0D0
!
      DO 240 L = 1, N
	 J = 0
	 H = MACHEP * (DABS(D(L)) + DABS(E(L)))
	 IF (B .LT. H) B = H
!
	 DO 110 M = L, N
	    IF (DABS(E(M)) .LE. B) GO TO 120
!
  110    CONTINUE
!
  120    IF (M .EQ. L) GO TO 220
  130    IF (J .EQ. 30) GO TO 1000
	 J = J + 1
!
	 L1 = L + 1
	 G = D(L)
	 P = (D(L1) - G) / (2.0D0 * E(L))
	 R = DSQRT(P*P+1.0D0)
	 D(L) = E(L) / (P + DSIGN(R,P))
	 H = G - D(L)
!
	 DO 140 I = L1, N
  140    D(I) = D(I) - H
!
	 F = F + H
!
	 P = D(M)
	 C = 1.0D0
	 S = 0.0D0
	 MML = M - L
!
	 DO 200 II = 1, MML
	    I = M - II
	    G = C * E(I)
	    H = C * P
	    IF (DABS(P) .LT. DABS(E(I))) GO TO 150
	    C = E(I) / P
	    R = DSQRT(C*C+1.0D0)
	    E(I+1) = S * P * R
	    S = C / R
	    C = 1.0D0 / R
	    GO TO 160
  150       C = P / E(I)
	    R = DSQRT(C*C+1.0D0)
	    E(I+1) = S * E(I) * R
	    S = 1.0D0 / R
	    C = C * S
  160       P = C * D(I) - S * G
	    D(I+1) = H + S * (C * G + S * D(I))
!
	    DO 180 K = 1, N
	       H = Z(K,I+1)
	       Z(K,I+1) = S * Z(K,I) + C * H
	       Z(K,I) = C * Z(K,I) - S * H
  180       CONTINUE
!
  200    CONTINUE
!
	 E(L) = S * P
	 D(L) = C * P
	 IF (DABS(E(L)) .GT. B) GO TO 130
  220    D(L) = D(L) + F
  240 CONTINUE
!
      DO 300 II = 2, N
	 I = II - 1
	 K = I
	 P = D(I)
!
	 DO 260 J = II, N
	    IF (D(J) .GE. P) GO TO 260
	    K = J
	    P = D(J)
  260    CONTINUE
!
	 IF (K .EQ. I) GO TO 300
	 D(K) = D(I)
	 D(I) = P
!
	 DO 280 J = 1, N
	    P = Z(J,I)
	    Z(J,I) = Z(J,K)
	    Z(J,K) = P
  280    CONTINUE
!
  300 CONTINUE
!
      GO TO 1001
!
 1000 IERR = L
 1001 RETURN
!
    END
!
!
!
! = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
! @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
    
! NR1989 page 88
        SUBROUTINE SPLIN0(X,Y,N,Y2)
        PARAMETER (NMAX=50)
        DIMENSION X(N),Y(N),Y2(N),U(NMAX)

        Y2(1)=0.
        U(1)=0.

        DO      I=2,N-1
                SIG=(X(I)-X(I-1))/(X(I+1)-X(I-1))
                P=SIG*Y2(I-1)+2.
                Y2(I)=(SIG-1.)/P
                U(I)=(6.*((Y(I+1)-Y(I))/(X(I+1)-X(I))-(Y(I)-Y(I-1))/(X(I)-X(I-1)))/(X(I+1)-X(I-1))-SIG*U(I-1))/P
        END DO

        QN=0.
        UN=0.

        Y2(N)=(UN-QN*U(N-1))/(QN*Y2(N-1)+1.)

        DO      K=N-1,1,-1
                Y2(K)=Y2(K)*Y2(K+1)+U(K)
        END DO

        RETURN
    END

! @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
    
! NR1989 page 89
      SUBROUTINE SPLINT(XA,YA,Y2A,N,X,Y)
      DIMENSION XA(N),YA(N),Y2A(N)
      KLO=1
      KHI=N
1     IF (KHI-KLO.GT.1) THEN
        K=(KHI+KLO)/2
        IF(XA(K).GT.X)THEN
          KHI=K
        ELSE
          KLO=K
        ENDIF
      GOTO 1
      ENDIF
      H=XA(KHI)-XA(KLO)
      IF (H.EQ.0.) PAUSE 'Bad XA input.'
      A=(XA(KHI)-X)/H
      B=(X-XA(KLO))/H
      Y=A*YA(KLO)+B*YA(KHI)+((A**3-A)*Y2A(KLO)+(B**3-B)*Y2A(KHI))*(H**2)/6.
      RETURN
      END

! @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
