subroutine DLLDipolarTriplet12NBimodial (lvector,dval,eval,ga,gb,rvector,rdistance,breala,bimaga,zmreal,zmimag,zfsmatrix)

  ! DLLDipolarTriplet fills up the UPPER TRIANGULAR PART of the energy matrix for a system with two S=1 spins 
  ! subject to dipolar interaction.
  ! The math and format/algoritm is described in BOOK 23, pp 59-75.
  ! @ WRHagen 20220723, 20221102
  ! DLLDipolarTriplet12N calculates the dipolar interaction of a central S=1 CuAc dimer with its 12 nearest
  ! neighbour CuAc dimers [book 24, pp 62-63].
  ! @ WRHagen 20221102
  ! The dipolar interaction strenght is bimodially distributed over triplet and singulet neighbours
  ! @ WRHagen 20221116; corrected 20221124

  ! History:    DLLDIP8CUBEN    20210120
  !             DLLDipolarNew   20220525    [book 23, pp 13-23]

 
IMPLICIT NONE
  !DEC$ ATTRIBUTES DLLEXPORT::DLLDipolarTriplet12NBimodial
  !DEC$ ATTRIBUTES STDCALL::DLLDipolarTriplet12NBimodial

  ! Declare calling parameters
    REAL*8, INTENT(IN) ::  lvector(3),dval,eval,ga(3),gb(3,3),rvector(3),rdistance
    REAL*8, INTENT(OUT) :: breala(9,9),bimaga(9,9),zmreal(9,9),zmimag(9,9),zfsmatrix(9,9)
    
  ! Declare local parameters
    INTEGER i,j,k
    REAL*8  sq2,beta,lx,ly,lz,gxa,gya,gza
    REAL*8  dreal1(9,9),dimag1(9,9),dreal2(9,9),dimag2(9,9)
    REAL*8  ax,ay,az,bx,by,bz,SIfactor
    REAL*8  rand

  ! Body of DLLDipolarTriplet
    sq2 = SQRT(2.)
    beta = 0.46689088 / 10000.  !in reciprocal cm per gauss
    lx = lvector(1)
    ly = lvector(2)
    lz = lvector(3)
    
    FORALL (i=1:9,j=1:9)
        breala(i,j) = 0.0
        bimaga(i,j) = 0.0
        zfsmatrix(i,j) = 0.0
        dreal1(i,j) = 0.0
        dimag1(i,j) = 0.0
        dreal2(i,j) = 0.0
        dimag2(i,j) = 0.0
        zmreal(i,j) = 0.0
        zmimag(i,j) = 0.0
    END FORALL
  
  ! Zero field splittings matrix (upper triangle)
    zfsmatrix(1,1) = dval/3.
    zfsmatrix(2,2) = zfsmatrix(1,1)
    zfsmatrix(3,3) = zfsmatrix(1,1)
    zfsmatrix(4,4) = -2.*dval / 3.
    zfsmatrix(5,5) = zfsmatrix(4,4)
    zfsmatrix(6,6) = zfsmatrix(4,4)
    zfsmatrix(7,7) = zfsmatrix(1,1)
    zfsmatrix(8,8) = zfsmatrix(1,1)
    zfsmatrix(9,9) = zfsmatrix(1,1)
    zfsmatrix(1,7) = eval
    zfsmatrix(2,8) = zfsmatrix(1,7)
    zfsmatrix(3,9) = zfsmatrix(1,7)
    
  ! Zeeman terms for B = 1.000 (upper triangle)
    gxa = beta * ga(1) * lx / 2.
    gya = beta * ga(2) * ly / 2.
    gza = beta * ga(3) * lz / 2.

    breala(1,1) = 2.*gza
    breala(1,4) = sq2*gxa
    breala(2,2) = breala(1,1)
    breala(2,5) = breala(1,4)
    breala(3,3) = breala(1,1)
    breala(3,6) = breala(1,4)
    breala(4,7) = breala(1,4)
    breala(5,8) = breala(1,4)
    breala(6,9) = breala(1,4)
    breala(7,7) = -breala(1,1)
    breala(8,8) = -breala(1,1)
    breala(9,9) = -breala(1,1)
    
    bimaga(1,4) = -sq2*gya
    bimaga(2,5) = bimaga(1,4)
    bimaga(3,6) = bimaga(1,4)
    bimaga(4,7) = bimaga(1,4)
    bimaga(5,8) = bimaga(1,4)
    bimaga(6,9) = bimaga(1,4)
    
  ! Now enter the 12 nearest-neighbour loop
  !  DO  k=1,12
  !      CALL RANDOM_SEED()              !added on 20221106
  !      CALL RANDOM_NUMBER(rand)
  !      IF(rand .gt. tripopul) CYCLE
        
  !      FORALL (i=1:3, j=1:3)
  !          gb(i,j) = gbIN(k,i,j)
  !      END FORALL
  !      rdistance = rdistanceIN(k)
  !      DO  i=1,3
  !          rvector(i) = rvectorIN(k,i) * rdistance
  !      END DO
        
    
  ! First dipolar term (upper triangle)
    dreal1(1,1) =  ga(3) * gb(3,3)
    dreal1(1,2) = 0.5*sq2*ga(3)*gb(3,1)
    dreal1(1,4) = 0.5*sq2*ga(1)*gb(1,3)
    dreal1(1,5) =  0.5*ga(1)*gb(1,1) - 0.5*ga(2)*gb(2,2)
    dreal1(2,4) =  0.5*ga(1)*gb(1,1) + 0.5*ga(2)*gb(2,2)
    dreal1(2,6) = dreal1(1,5)
    dreal1(3,3) = -dreal1(1,1)
    dreal1(3,5) =  dreal1(2,4)
    dreal1(3,6) = -dreal1(1,4)
    dreal1(4,7) = dreal1(1,4)
    dreal1(4,8) = dreal1(1,5)
    dreal1(5,7) = dreal1(2,4)
    dreal1(5,9) = dreal1(1,5)
    dreal1(6,8) = dreal1(2,4)
    dreal1(6,9) = -dreal1(1,4)
    dreal1(7,7) = -dreal1(1,1)
    dreal1(7,8) = -dreal1(1,2)
    dreal1(8,9) = -dreal1(1,2)
    dreal1(9,9) = dreal1(1,1)
   
    dimag1(1,2) = -0.5*sq2*ga(3)*gb(3,2)
    dimag1(1,4) = -0.5*sq2*ga(2)*gb(2,3)
    dimag1(1,5) = -0.5*ga(1)*gb(1,2) - 0.5*ga(2)*gb(2,1)
    dimag1(2,4) = 0.5*ga(1)*gb(1,2) - 0.5*ga(2)*gb(2,1)
    dimag1(2,6) = dimag1(1,5)      !-
    dimag1(3,5) = dimag1(2,4)
    dimag1(3,6) = -dimag1(1,4)
    dimag1(4,7) = dimag1(1,4)
    dimag1(4,8) = dimag1(1,5)
    dimag1(5,7) = dimag1(2,4)
    dimag1(5,9) = dimag1(1,5)
    dimag1(6,8) = dimag1(2,4)
    dimag1(6,9) = -dimag1(1,4)
    dimag1(7,8) = -dimag1(1,2)
    dimag1(8,9) = -dimag1(1,2)
    
  ! Second dipolar term (upper triangle)
    ax = rvector(1) * ga(1)
    ay = rvector(2) * ga(2)
    az = rvector(3) * ga(3)
    bx = rvector(1)*gb(1,1) + rvector(2)*gb(2,1) + rvector(3)*gb(3,1)
    by = rvector(1)*gb(1,2) + rvector(2)*gb(2,2) + rvector(3)*gb(3,2)
    bz = rvector(1)*gb(1,3) + rvector(2)*gb(2,3) + rvector(3)*gb(3,3)
    
    dreal2(1,1) =az*bz
    dreal2(1,2) = 0.5*sq2*az*bx
    dreal2(1,4) = 0.5*sq2*ax*bz
    dreal2(1,5) =0.5*ax*bx - 0.5*ay*by
    dreal2(2,4) = 0.5*ax*bx + 0.5*ay*by
    dreal2(2,6) = dreal2(1,5)
    dreal2(3,3) = -dreal2(1,1)
    dreal2(3,5) =  dreal2(2,4)
    dreal2(3,6) = -dreal2(1,4)
    dreal2(4,7) = dreal2(1,4)
    dreal2(4,8) = dreal2(1,5)
    dreal2(5,7) = dreal2(2,4)
    dreal2(5,9) = dreal2(1,5)
    dreal2(6,8) = dreal2(2,4)
    dreal2(6,9) = -dreal2(1,4)
    dreal2(7,7) = -dreal2(1,1)
    dreal2(7,8) = -dreal2(1,2)
    dreal2(8,9) = -dreal2(1,2)
    dreal2(9,9) = dreal2(1,1)
    
    dimag2(1,2) = -0.5*sq2*az*by
    dimag2(1,4) = -0.5*sq2*ay*bz
    dimag2(1,5) = -0.5*ax*by - 0.5*ay*bx
    dimag2(2,4) = 0.5*ax*by - 0.5*ay*bx
    dimag2(2,6) = dimag2(1,5)
    dimag2(3,5) = dimag2(2,4)
    dimag2(3,6) = -dimag2(1,4)
    dimag2(4,7) = dimag2(1,4)
    dimag2(4,8) = dimag2(1,5)
    dimag2(5,7) = dimag2(2,4)
    dimag2(5,9) = dimag2(1,5)
    dimag2(6,8) = dimag2(2,4)
    dimag2(6,9) = -dimag2(1,4)
    dimag2(7,8) = -dimag2(1,2)
    dimag2(8,9) = -dimag2(1,2)
    
    dreal2 = -3.0 * dreal2 / (rdistance*rdistance)
    dimag2 = -3.0 * dimag2 / (rdistance*rdistance)
    
  ! Next lines taken from DLLDIP8CUBEN:
  ! Note that in the point-dipole equation (book 13, page 47):
  ! the first beta is in wavenumbers/gauss since hv is calculated in wavenumbers;
  ! and the second beta is in SI units: joule/tesla.
  ! The pre-factor of the dipolar term is mu-zero x beta-squared / 4 pi x r^3,
  ! wherin beta-squared is first beta (wavenumbers/gauss) times second beta (joule/tesla).
  ! The pre-factor is first-beta x SIfactor.
    
  ! SIfactor is (mu-nul/4pi) * (beta/r**3) with ALL quantities in SI units
  ! and for F in Hertz (not MHz) and rdistance from cm (not m) to Angstrom 
    SIfactor = 9.274009994 * 1000.   
    zmreal = zmreal + beta * SIfactor * (dreal1 + dreal2) / (rdistance*rdistance*rdistance)
    zmimag = zmimag + beta * SIfactor * (dimag1 + dimag2) / (rdistance*rdistance*rdistance)
    DO  j=2,9
        DO  i=1,j-1
            zmreal(j,i) = zmreal(i,j)
            zmimag(j,i) = zmimag(i,j)
        END DO
    END DO

  !  END DO  ! End of 12-neighbour loop
    
  ! Correct (= reduce) for fractional triplet population as a function of temperature [20221101]
  !  zmreal = zmreal * tripopul                 ! deleted on 20221106
  !  zmimag = zmimag * tripopul
    
  ! Now add the zero-field splittings:
  !  zmreal = zmreal + zfsmatrix
    
  ! Add lower triangular part for EISPACK routines [20221031]
    DO  j=2,9
        DO  i=1,j-1
            breala(j,i) = breala(i,j)
            bimaga(j,i) = bimaga(i,j)
            zfsmatrix(j,i) = zfsmatrix(i,j)
        END DO
    END DO
    
end subroutine DLLDipolarTriplet12NBimodial
